/*
 nasm -f elf busasm.asm   for busasm.o 
 gcc busspeedmp.c cpuidc.c  busasm.o -lrt -lc -lm -O3 -o MPbusspeed32
 ./MPbusspeed32

 nasm -f elf64 -g -F stabs busasm64.asm   for busasm64.o 
 gcc busspeedmp.c cpuidc64.c  busasm64.o -m64 -lrt -lc -lm -O3 -o MPbusspeed64
 ./MPbusspeed64

 *  BusSpeedMP
 *
 */

 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include "cpuidh.h"
 #include <math.h>
 #include <inttypes.h>
 #include <stdint.h>
 #include <malloc.h>
 #include <mm_malloc.h>


//  ************** PROTOPTYPES *****************

void inc32words(int t);
void inc16words(int t);
void inc8words(int t);
void inc4words(int t);
void inc2words(int t);
void inc1word(int t);
void incSSE2(int t);
void adjust();


// ************* GLOBAL VARIABLES **********

//  #define Bits64
 #define Bits32



#ifdef Bits64
 long long * array;
 long long    andsum1[64];
 long long    andsumx = 0x5555555555555555;
 unsigned int wordBytes = 8;
 char   version[30] = "64 bit Version 1.0,";
 char mips[80] = " 64 bit words, Speed in MB/Second - MIPS divide by 8"; 
#else
 int        * array;
 int          andsum1[64];
 int          andsumx = 0x55555555;
 unsigned int wordBytes = 4;
 char   version[30] = "32 bit Version 1.0,";
 char mips[80] = " 32 bit words, Speed in MB/Second - MIPS divide by 4"; 
#endif

int           * andarray;
int            andall[256];
int            sse2loopinc;
long long    * arraym;
long long    * andSSE2;
 
 volatile int wordsToTest;
 volatile int passes1;
 volatile int testToRun;


 unsigned int memoryKBytes[20];
 unsigned int ramKB;
 unsigned int ramMB;
 int           ipass;
 int           threads = 1;
 int           memoryWords[20];
 int           loopx;
 int           runStage = -2;
 int           runStage2 = 0;
 int           runPasses = 10;

 double        memoryMB[20];
 double        runSecs = 0.05;

 pthread_t tid[100]; 
 pthread_attr_t * attrt = NULL; 

 FILE          *outfile;

void *runTests(void *arg)
{
     long thread;

     thread = (long)arg;
     
     if (testToRun == 1) inc32words(thread);   
     if (testToRun == 2) inc16words(thread);
     if (testToRun == 3) inc8words(thread);
     if (testToRun == 4) inc4words(thread);
     if (testToRun == 5) inc2words(thread);
     if (testToRun == 6) inc1word(thread);
     if (testToRun == 7) incSSE2(thread);   
     return;
}

int main(int argc, char *argv[])
{
   int        i, j, g, p;
   int        endTest;
   int        param01 = 0;
   int        mbpersec;
   long       t;
   double     addressInc[10];
   char       writeMsg[200];

    for (i=1; i<3; i=i+2)
    {
       if (argc > i)
       {
          switch (toupper(argv[i][0]))
          {
                case 'T':
                if (argc > i+1)
                {
                   sscanf(argv[i+1],"%d", &param01);
                   if (param01 > 0) threads = param01;
                }
                break;
         }
       }
    }

   addressInc[1] = 32; // inc32words
   addressInc[2] = 16; // inc16words
   addressInc[3] = 8;  // inc8words
   addressInc[4] = 4;  // inc4words
   addressInc[5] = 2;  // inc2words
   addressInc[6] = 1;  // inc1word
   addressInc[7] = 1;  // inc all words incSSE2

   memoryKBytes[0] = 6;
   memoryKBytes[1] = 24;
   memoryKBytes[2] = 96;
   memoryKBytes[3] = 384;
   memoryKBytes[4] = 768;
   memoryKBytes[5] = 1536;
   memoryKBytes[6] = 16380;  //  16384;
   memoryKBytes[7] = 131070; // 131072;
   memoryKBytes[8] = 393210; // 393216

   for (i=0;i<64;i++)
   {
       andsum1[i] = andsumx;
   }

   local_time();
   outfile = fopen("busSpeedMP.txt","a+");
   if (outfile == NULL)
   {
        printf ("Cannot open results file \n\n");
        printf(" Press Enter\n");
        g  = getchar();
        exit (0);
   }
   getDetails();
   if (param01 == 0)
   {
       threads = CPUavail;
   }
   if (threads > 32)
   {
       threads = 64;
   }
   else if (threads > 16)
   {
       threads = 32;
   }
   else if (threads >  8)
   {
       threads = 16;
   }
   else if (threads >  4)
   {
       threads =  8;
   }
   else if (threads >  2)
   {
       threads =  4;
   }
   printf("\n");
   for (i=1; i<10; i++)
   {
       printf("%s\n", configdata[i]);
   }
   fprintf (outfile, " #####################################################################\n\n");
   for (i=1; i<10; i++)
   {
       fprintf(outfile, "%s \n", configdata[i]);
   }
   printf("\n");
   printf(" #############################################################\n\n"); 
   printf ("  MP Bus Speeds %s %d Threads, %s\n", version, threads, timeday);
   printf ("   Kbytes Inc32wds Inc16wds  Inc8wds  Inc4wds  Inc2wds  ReadAll 128bSSE2\n\n");


   fprintf (outfile, "\n #####################################################################\n\n");
   fprintf (outfile, "  MP Bus Speeds %s %d Threads, %s\n", version, threads, timeday);

   fprintf (outfile, "   Kbytes Inc32wds Inc16wds  Inc8wds  Inc4wds  Inc2wds  ReadAll 128bSSE2\n\n");


   ipass = 9;
   ramKB = (unsigned int)(ramGB * 1000000);
   ramMB = ramKB / 1024;
   if (ramMB < 1024) ipass = 8; 

   if (ramMB < 384 ) ipass = 7;
   if (ramMB < 32  ) ipass = 6;
   for (i=0; i<ipass; i++)
   {
       memoryWords[i]  = (int)((double)memoryKBytes[i] * 1024.0/ (double)wordBytes) ;
       memoryMB[i] = ((double)memoryKBytes[i] * 1024.0)/ 1000000;        
   }


#ifdef Bits64
   array = (long long *)_mm_malloc(memoryKBytes[ipass-1]*1024, 16);
   if (array == NULL)
   {
       printf(" Cannot allocate memory ERROR WILL EXIT\n");
       printf(" Press Enter\n");
       g  = getchar();
       exit(1);
   }
#else
   array = (int *)_mm_malloc(memoryKBytes[ipass-1]*1024, 16);
   if (array == NULL)
   {
       printf(" Cannot allocate memory ERROR WILL EXIT\n");
       printf(" Press Enter\n");
       g  = getchar();
       exit(1);
   }
 #endif
   arraym  = (long long  *)array;

   andarray = (int *)_mm_malloc(1024, 16);
   if (andarray == NULL)
   {
       printf(" Cannot allocate memory ERROR WILL EXIT\n");
       printf(" Press Enter\n");
       _mm_free(array);
       g  = getchar();
       exit(1);
   }
   andSSE2 = (long long  *)andarray; 

   for (i=0; i<memoryWords[ipass-1]; i++) // arraym
   {
       array[i] = andsumx;
   }


   for (j=0; j<256; j++) // andSSE2
   {
       andall[j] = 0x55555555;
   }
   for (j=0; j<4; j++) // andSSE2
   {
       andarray[j] = 0x55555555;
   }

   
   runStage = 0;
   runStage2 = 0;

   endTest = 7;
   if (hasSSE2) endTest = 8;

   for (i=0; i<ipass; i++)
   {
       wordsToTest = memoryWords[i];
       runStage = runStage + 1;
       printf(" %8d", memoryKBytes[i]);
       fprintf(outfile, " %8d", memoryKBytes[i]);
       for (testToRun=1; testToRun<endTest; testToRun++)
       {
           passes1 = 1;
           for (loopx=0; loopx<10; loopx++)
           {
               start_time();
               for (t=0; t<threads; t++)
               {
                    pthread_create(&tid[t], attrt, runTests, (void *)t);
               }
               for (t=0; t<threads; t++)
               {
                   pthread_join(tid[t], NULL);                
               }
               end_time();
               adjust();
           }
      
           start_time();
           for (p=0; p<runPasses; p++)
           {
               for (t=0; t<threads; t++)
               {
                    pthread_create(&tid[t], attrt, runTests, (void *)t);
               }
               for (t=0; t<threads; t++)
               {
                   pthread_join(tid[t], NULL);                
               }
           }
           end_time();
           mbpersec = (int)((double)threads * (double)passes1 * (double)runPasses * memoryMB[i] / secs / addressInc[testToRun]);
           printf(" %8d", mbpersec);           
           fprintf(outfile," %8d", mbpersec);           
           fflush(stdout);
           fflush(outfile);
       }
       printf("\n");   
       fprintf(outfile, "\n");   
       for (t=0; t<threads; t++)
       {

           if (andsum1[t] != andsumx)
           {
              #ifdef Bits64
               sprintf(writeMsg, " Wrong result Expected 0x%16.16lx Was 0x%16.16lx at %d KB Thread %ld", (long unsigned int)andsumx, (long unsigned int)andsum1[t], memoryKBytes[i], t+1);
              #else
               sprintf(writeMsg, " Wrong result Expected 0x%8.8x Was 0x%8.8x at %d KB Thread %ld", andsumx, andsum1[t], memoryKBytes[i], t+1);
              #endif
               printf(" %s\n", writeMsg);
               fprintf(outfile, " %s\n", writeMsg);
               andsum1[t] = andsumx;
           }
           if (hasSSE2)
           {
               for (j=0; j<4; j++) // andSSE2
               {
                   if (andall[j+t*4]  != 0x55555555)
                   {
                       sprintf(writeMsg, " Wrong result Expected 0x55555555 Was 0x%8.8x at %d KB SSE2 Thread %ld", andall[j+t*4], memoryKBytes[i], t+1);
                       printf(" %s\n", writeMsg);
                       fprintf(outfile, " %s\n", writeMsg);
                      andarray[j] = 0x55555555;
                   }
               }    
           }
       }
   }
   printf("\n");   
   fprintf(outfile, "\n");   

   local_time();
   printf ("           %s\n\n", mips);
   printf ("                   End at %s\n", timeday);
   fprintf (outfile, "           %s\n\n", mips);
   fprintf (outfile, "                   End at %s\n", timeday);

   fprintf(outfile, "\n");
   fflush (outfile);
   fclose(outfile);
   _mm_free(andarray);  
   _mm_free(array);

   return 1;
}

void inc32words(int t)
{
   int i, j;

   for(j=0; j<passes1; j++)
   {
       for (i=0; i<wordsToTest; i=i+768)
       {
           andsum1[t] = andsum1[t] & array[i     ] & array[i+32  ] & array[i+64  ] & array[i+96  ]
                                   & array[i+128 ] & array[i+160 ] & array[i+192 ] & array[i+224 ]
                                   & array[i+256 ] & array[i+288 ] & array[i+320 ] & array[i+352 ]
                                   & array[i+384 ] & array[i+416 ] & array[i+448 ] & array[i+480 ]
                                   & array[i+512 ] & array[i+544 ] & array[i+576 ] & array[i+608 ]
                                   & array[i+640 ] & array[i+672 ] & array[i+704 ] & array[i+736 ];
       }
   }
}

void inc16words(int t)
{
   int i, j;

   for(j=0; j<passes1; j++)
   {
       for (i=0; i<wordsToTest; i=i+768)
       {
           andsum1[t] = andsum1[t] & array[i    ] & array[i+16 ] & array[i+32 ] & array[i+48 ]
                                   & array[i+64 ] & array[i+80 ] & array[i+96 ] & array[i+112]
                                   & array[i+128] & array[i+144] & array[i+160] & array[i+176]
                                   & array[i+192] & array[i+208] & array[i+224] & array[i+240]
                                   & array[i+256] & array[i+272] & array[i+288] & array[i+304]
                                   & array[i+320] & array[i+336] & array[i+352] & array[i+368]
                                   & array[i+384] & array[i+400] & array[i+416] & array[i+432]
                                   & array[i+448] & array[i+464] & array[i+480] & array[i+496]
                                   & array[i+512] & array[i+528] & array[i+544] & array[i+560]
                                   & array[i+576] & array[i+592] & array[i+608] & array[i+624]
                                   & array[i+640] & array[i+656] & array[i+672] & array[i+688]
                                   & array[i+704] & array[i+720] & array[i+736] & array[i+752];
       }
   }
}

void inc8words(int t)
{
   int i, j;

   for(j=0; j<passes1; j++)
   {
       for (i=0; i<wordsToTest; i=i+384)
       {
           andsum1[t] = andsum1[t] & array[i    ] & array[i+8  ] & array[i+16 ] & array[i+24 ]
                                   & array[i+32 ] & array[i+40 ] & array[i+48 ] & array[i+56 ]
                                   & array[i+64 ] & array[i+72 ] & array[i+80 ] & array[i+88 ]
                                   & array[i+96 ] & array[i+104] & array[i+112] & array[i+120]
                                   & array[i+128] & array[i+136] & array[i+144] & array[i+152]
                                   & array[i+160] & array[i+168] & array[i+176] & array[i+184]
                                   & array[i+192] & array[i+200] & array[i+208] & array[i+216]
                                   & array[i+224] & array[i+232] & array[i+240] & array[i+248]
                                   & array[i+256] & array[i+264] & array[i+272] & array[i+280]
                                   & array[i+288] & array[i+296] & array[i+304] & array[i+312]
                                   & array[i+320] & array[i+328] & array[i+336] & array[i+344]
                                   & array[i+352] & array[i+360] & array[i+368] & array[i+376];
       }
   }
}

void inc4words(int t)
{
   int i, j;

   for(j=0; j<passes1; j++)
   {
       for (i=0; i<wordsToTest; i=i+256)
       {
           andsum1[t] = andsum1[t] & array[i    ] & array[i+4  ] & array[i+8  ] & array[i+12 ]
                                   & array[i+16 ] & array[i+20 ] & array[i+24 ] & array[i+28 ]
                                   & array[i+32 ] & array[i+36 ] & array[i+40 ] & array[i+44 ]
                                   & array[i+48 ] & array[i+52 ] & array[i+56 ] & array[i+60 ]
                                   & array[i+64 ] & array[i+68 ] & array[i+72 ] & array[i+76 ]
                                   & array[i+80 ] & array[i+84 ] & array[i+88 ] & array[i+92 ]
                                   & array[i+96 ] & array[i+100] & array[i+104] & array[i+108]
                                   & array[i+112] & array[i+116] & array[i+120] & array[i+124]
                                   & array[i+128] & array[i+132] & array[i+136] & array[i+140]
                                   & array[i+144] & array[i+148] & array[i+152] & array[i+156]
                                   & array[i+160] & array[i+164] & array[i+168] & array[i+172]
                                   & array[i+176] & array[i+180] & array[i+184] & array[i+188]
                                   & array[i+192] & array[i+196] & array[i+200] & array[i+204]
                                   & array[i+208] & array[i+212] & array[i+216] & array[i+220]
                                   & array[i+224] & array[i+228] & array[i+232] & array[i+236]
                                   & array[i+240] & array[i+244] & array[i+248] & array[i+252];
       }
   }
}

void inc2words(int t)
{
   int i, j;

   for(j=0; j<passes1; j++)
   {
       for (i=0; i<wordsToTest; i=i+128)
       {
           andsum1[t] = andsum1[t] & array[i    ] & array[i+2  ] & array[i+4  ] & array[i+6  ]
                                   & array[i+8  ] & array[i+10 ] & array[i+12 ] & array[i+14 ]
                                   & array[i+16 ] & array[i+18 ] & array[i+20 ] & array[i+22 ]
                                   & array[i+24 ] & array[i+26 ] & array[i+28 ] & array[i+30 ]
                                   & array[i+32 ] & array[i+34 ] & array[i+36 ] & array[i+38 ]
                                   & array[i+40 ] & array[i+42 ] & array[i+44 ] & array[i+46 ]
                                   & array[i+48 ] & array[i+50 ] & array[i+52 ] & array[i+54 ]
                                   & array[i+56 ] & array[i+58 ] & array[i+60 ] & array[i+62 ]
                                   & array[i+64 ] & array[i+66 ] & array[i+68 ] & array[i+70 ]
                                   & array[i+72 ] & array[i+74 ] & array[i+76 ] & array[i+78 ]
                                   & array[i+80 ] & array[i+82 ] & array[i+84 ] & array[i+86 ]
                                   & array[i+88 ] & array[i+90 ] & array[i+92 ] & array[i+94 ]
                                   & array[i+96 ] & array[i+98 ] & array[i+100] & array[i+102]
                                   & array[i+104] & array[i+106] & array[i+108] & array[i+110]
                                   & array[i+112] & array[i+114] & array[i+116] & array[i+118]
                                   & array[i+120] & array[i+122] & array[i+124] & array[i+126];
       }
    }
}

void inc1word(int t)
{
   int i, j;

   for(j=0; j<passes1; j++)
   {
       for (i=0; i<wordsToTest; i=i+64)
       {
           andsum1[t] = andsum1[t] & array[i   ] & array[i+1 ] & array[i+2 ] & array[i+3 ]
                                   & array[i+4 ] & array[i+5 ] & array[i+6 ] & array[i+7 ]
                                   & array[i+8 ] & array[i+9 ] & array[i+10] & array[i+11]
                                   & array[i+12] & array[i+13] & array[i+14] & array[i+15]
                                   & array[i+16] & array[i+17] & array[i+18] & array[i+19]
                                   & array[i+20] & array[i+21] & array[i+22] & array[i+23]
                                   & array[i+24] & array[i+25] & array[i+26] & array[i+27]
                                   & array[i+28] & array[i+29] & array[i+30] & array[i+31]
                                   & array[i+32] & array[i+33] & array[i+34] & array[i+35]
                                   & array[i+36] & array[i+37] & array[i+38] & array[i+39]
                                   & array[i+40] & array[i+41] & array[i+42] & array[i+43]
                                   & array[i+44] & array[i+45] & array[i+46] & array[i+47]
                                   & array[i+48] & array[i+49] & array[i+50] & array[i+51]
                                   & array[i+52] & array[i+53] & array[i+54] & array[i+55]
                                   & array[i+56] & array[i+57] & array[i+58] & array[i+59]
                                   & array[i+60] & array[i+61] & array[i+62] & array[i+63];
       }
   }
}

void incSSE2(int t)
{
   int j, tt;
   tt = t * 4;
   sse2loopinc = wordsToTest * wordBytes / 1024;
   for(j=0; j<passes1; j++)
   {
      _asmsse2();
   }
   andall[tt  ] = andall[tt  ] & andarray[0];
   andall[tt+1] = andall[tt+1] & andarray[1];
   andall[tt+2] = andall[tt+2] & andarray[2];
   andall[tt+3] = andall[tt+3] & andarray[3];  
}

void adjust()
{
   double runtime = runSecs;
   
   if (secs > runtime)
   {
       loopx = 11;
   }
   else if (secs < runtime / 100)
   {
       passes1 = passes1 * 100;
   }
   else if (secs > runtime / 2)
   {
       passes1 = passes1 * (int)(runtime / secs);
       loopx = 11;
   }
   else
   {
       passes1 = passes1 * (int)(runtime / secs);
   }
}
