/*
  nasm -f elf64 -g -F stabs asscode64.asm   for asscode64.o
  nasm -f elf  asscode32.asm                for asscode32.o
  
  gcc cpumaxmp.c cpuidc.c asscode64.o -m64 -lc -lm -lrt -o cpumaxmp64
  ./cpumaxmp64
  gcc cpumaxmp.c cpuidc.c asscode32.o -lc -lm -lrt -o cpumaxmp32
  ./cpumaxmp32

*/

  #include <stdio.h> 
  #include "cpuidh.h"
  #include <stdlib.h> 
  #include <unistd.h>   
  #include <math.h>
  #include <inttypes.h>
  #include <stdint.h>
  #include <malloc.h>
  #include <mm_malloc.h>
  #include <pthread.h> 

//  #define Bits64
  #define Bits32

  #ifdef Bits64
  #define version  "64 bit Version 1.0"
  #else
  #define version  "32 bit Version 1.0"
  #endif
  
#ifdef Bits64
 typedef struct 
 {
    float      *adsum;
    float      *assum1;
    long long  icounta;
 }
 MPDATA;
 
 long long  intCounta;
#else
 typedef struct 
 {
    float      *adsum;
    float      *assum1;
    unsigned int  icounta;
 }
 MPDATA;
 
 unsigned int  intCounta;
#endif



 MPDATA mem1;
 int mipst[100];
 int mflopst[100];

 FILE *outfile;
 
 float  * dsum;
 float  * ssum1;

 double totalS1 = 8000000000.0;

 pthread_t tid[100]; 
 pthread_attr_t * attrt = NULL; 
 pthread_mutex_t mutext = PTHREAD_MUTEX_INITIALIZER;

 void *runIntTests(void *arg)
 {   
    double secsHere;

    int    i, g;
    long   offset;
    int    mips = 0;

    offset = (long)arg;

    intCounta = mem1.icounta;
#ifdef Bits64
   long long  totalI1a = 2200000000;

   _mipsa4Reg();
   pthread_mutex_lock( &mutext);
   end_time();
   secsHere = secs;
   mips = (int)((double)intCounta / 1000000.0 / secsHere + 0.5);
   printf(" Thread %3ld - %7d 64 Bit Integer MIPS\n", offset+1, mips);
   fprintf(outfile, " Thread %3ld - %7d 64 bit Integer MIPS\n", offset+1, mips);
   mipst[offset] = mips;                                     
   if(intCounta != totalI1a)
   {
       fprintf(outfile, " 64 bit MIPS ADD ERRORS was %lld should be %lld\n", intCounta, totalI1a);
       printf(" 64 bit MIPS ADD ERRORS was %lld should be %lld\n", intCounta, totalI1a);
   }
   pthread_mutex_unlock( &mutext);
#else
   unsigned int  totalI1a = 2200000000;

   _mipsa4Reg();
   pthread_mutex_lock( &mutext);
   end_time();
   secsHere = secs;
   mips = (int)((double)intCounta / 1000000.0 / secsHere + 0.5);
   printf(" Thread %3ld - %7d 32 Bit Integer MIPS\n", offset+1, mips);
   fprintf(outfile, " Thread %3ld - %7d 32 bit Integer MIPS\n", offset+1, mips);
   mipst[offset] = mips;                                     
   if(intCounta != totalI1a)
   {
       fprintf(outfile, " 32 bit MIPS ADD ERRORS was %d should be %d\n", intCounta, totalI1a);
       printf(" 32 bit MIPS ADD ERRORS was %d should be %d\n", intCounta, totalI1a);
   }
   pthread_mutex_unlock( &mutext);
#endif
 }

 void *runSSETests(void *arg)
 {   
   double sseSum       = 2.684354560e+08;
   double SSEcount     = 0;
   double secsHere;

   int    i, g;
   int    mflops = 0;
   long   offset;
   float  *assumx;

   offset = (long)arg;
   dsum      =    mem1.adsum   + offset * 4;
   ssum1     =    mem1.assum1  + offset * 4;

   assumx = (float *)ssum1;
   if (hasSSE)
   {
       ssum1[4] = 1;
       ssum1[5] = 1;
       ssum1[6] = 1;
       ssum1[7] = 1;
       ssum1[0] = 0;
       ssum1[1] = 0;
       ssum1[2] = 0;
       ssum1[3] = 0;

       _SSE4Reg();
       pthread_mutex_lock( &mutext);
       end_time();
       ssum1 = (float *)assumx;
       secsHere = secs;
       mflops = (int)(totalS1 / 1000000.0 / secsHere + 0.5);
       SSEcount = SSEcount + (double) ssum1[0] + (double) ssum1[1]
                           + (double) ssum1[2] + (double) ssum1[3];

       printf(" Thread %3ld - %7d 32 Bit SSE MFLOPS\n", offset+1, mflops);
       fprintf(outfile, " Thread %3ld - %7d 32 Bit SSE MFLOPS\n", offset+1, mflops);

       mflopst[offset] = mflops;                                     
       if(SSEcount != sseSum)
       {
           fprintf(outfile, " 32 bit SSE MFLOPS ADD ERRORS was %15.9e should be %15.9e\n", SSEcount, sseSum);
           printf(" 32 bit SSE MFLOPS ADD ERRORS was %15.9e should be %15.9e\n", SSEcount, sseSum);
       }
       pthread_mutex_unlock( &mutext);
   }
   else
   {
       fprintf(outfile, "\n SSE Instructions Not Available\n\n");
       printf("\n SSE Instructions Not Available\n\n");       
   }
 }
 
int main(int argc, char *argv[])
{
    int  i, g;
    long ii;
    int  threads = 4;
    int  mips;
    int  mflops;
    int  memSize;
    int  param1;
    float     *adsum;  // temp store 2 SSE registers - 256 bits or 32 bytes or 8 x 32 bit words
    float     *assum1; // SSE input and output       - 128 bits or 16 bytes or 8 x 32 bit words 

    for (i=1; i<3; i=i+2)
    {
       if (argc > i)
       {
          switch (toupper(argv[i][0]))
          {
               case 'T':
                param1 = 0;
                if (argc > i+1)
                {
                   sscanf(argv[i+1],"%d", &param1);
                   if (param1 > 0) threads = param1;
                   if (threads > 64) threads = 64;
                }
                break;
         }
       }
    }

    outfile = fopen("MPadds.txt","a+");
    if (outfile == NULL)
    {
       printf(" Cannot open log file\n");
       printf(" Press Enter\n");
       g  = getchar();
       exit(1);
    }
    memSize = 4096;
    assum1 = (float *)_mm_malloc(memSize, 16);
    if (assum1 == NULL)
    {
       printf(" ERROR WILL EXIT\n");
       printf(" Press Enter\n");
       g  = getchar();
       exit(1);
    }
    adsum = (float *)_mm_malloc(memSize, 16);
    if (adsum == NULL)
    {
       printf(" ERROR WILL EXIT\n");
       printf(" Press Enter\n");
       g  = getchar();
       exit(1);
    }

    mem1.adsum   = (float *)adsum;
    mem1.assum1  = (float *)assum1;
    mem1.icounta = 0;
    
    printf("\n\n");

    local_time();
    printf("\n  %s", timeday);

    getDetails();
    printf("\n");
    for (i=1; i<10; i++)
    {
        printf("%s\n", configdata[i]);
    }
    printf("\n");
    fprintf (outfile, "\n"); 
    fprintf (outfile, "##############################################\n\n");
    for (i=1; i<10; i++)
    {
        fprintf(outfile, "%s \n", configdata[i]);
    }
    fprintf (outfile, "\n %d CPUs Available\n", CPUavail);
    fprintf (outfile, "##############################################\n\n");
    fprintf(outfile, " Multithreading Add Test %s %s \n", version, timeday);
    fprintf(outfile, " Integer Additions %d Threads\n\n", threads);
    printf(" Multithreading Add Test %s %s \n", version, timeday);
    printf(" Integer Additions %d Threads\n\n", threads);

    mips = 0;
    
    start_time();
    for (ii=0; ii<threads; ii++)
    {
        pthread_create(&tid[ii], attrt, runIntTests, (void *)ii);
    }
    for (ii=0; ii<threads; ii++)
    {
        pthread_join(tid[ii], NULL);
        mips = mips + mipst[ii];      
    }
#ifdef Bits64
    printf(" Total      - %7d 64 Bit Integer MIPS\n", mips);
    fprintf(outfile, " Total      - %7d 64 Bit Integer MIPS\n", mips);   

    mips = (int)((double)intCounta * (double)threads / 1000000.0 / secs + 0.5);
    printf(" Aggregate  - %7d 64 Bit Integer MIPS, based on last to finish\n", mips);
    fprintf(outfile, " Aggregate  - %7d 64 Bit Integer MIPS, based on last to finish\n", mips);
#else
    printf(" Total      - %7d 32 Bit Integer MIPS\n", mips);
    fprintf(outfile, " Total      - %7d 32 Bit Integer MIPS\n", mips);   

    mips = (int)((double)intCounta * (double)threads / 1000000.0 / secs + 0.5);
    printf(" Aggregate  - %7d 32 Bit Integer MIPS, based on last to finish\n", mips);
    fprintf(outfile, " Aggregate  - %7d 32 Bit Integer MIPS, based on last to finish\n", mips);
#endif

    printf("\n\n SSE Floating Point Additions %d Threads\n\n", threads);
    fprintf(outfile, "\n\n SSE Floating Point Additions %d Threads\n\n", threads);

    mflops = 0;
    start_time();
    for (ii=0; ii<threads; ii++)
    {
        pthread_create(&tid[ii], attrt, runSSETests, (void *)ii);
    }
    for (ii=0; ii<threads; ii++)
    {
        pthread_join(tid[ii], NULL);
        mflops = mflops + mflopst[ii];      
    }
    printf(" Total      - %7d 32 Bit SSE MFLOPS\n", mflops);
    fprintf(outfile, " Total      - %7d 32 Bit SSE MFLOPS\n", mflops);
    mflops = (int)(totalS1 * (double)threads / 1000000.0 / secs + 0.5);
    printf(" Aggregate  - %7d 32 Bit SSE MFLOPS, based on last to finish\n", mflops);
    fprintf(outfile, " Aggregate  - %7d 32 Bit SSE MFLOPS, based on last to finish\n", mflops);

    local_time();
    printf("\n End of test %s", timeday);
    fprintf(outfile, "\n End of test %s", timeday);

    fprintf(outfile, "\n");
    fclose(outfile);
    _mm_free(adsum);
    _mm_free(assum1); 
    printf("\n Results also in log file MPadds.txt\n\n");
    printf(" Press Enter\n");
    g  = getchar();  
 
    return 0;
}

